/**
  Draw multiple objects in a single raytracer pass.
  
  Dr. Orion Sky Lawlor, olawlor@acm.org, 2012-01-30 (Public Domain)
*/
#include "physics/world.h" /* physics::library and physics::object */
#include "ogl/glsl.h"

/* Just include library bodies here, for easy linking */
#include "physics/world.cpp" 
#include "physics/config.cpp"
#include "ogl/glsl.cpp"


/* A raytraced object */
class rayObject : public physics::object {
public:
	rayObject(void) 
		:physics::object(0.01)  /* <- our timestep, in seconds */
	{ }
	
	void simulate(physics::library &lib) { }
	
	void draw(physics::library &lib) {
		/* Load up shader */
		GLhandleARB prog=MAKE_PROGRAM_OBJECT3(
		/* vertex program */
			void main(void) {
			   myColor = gl_Color;
			   /* Include everything in onscreen position */
			   gl_Position=gl_ModelViewProjectionMatrix*gl_Vertex;
			   /* Don't project, but do modelview transforms */
			   location=(vec3)(gl_ModelViewMatrix*gl_Vertex);
			}
		,
		/* shared vertex/fragment code */
			uniform vec3 camera; // world coordinates of camera
			varying vec4 myColor;
			varying vec3 location; // world coordinates of our pixel
		,
		/* fragment program */
			
			/* Return the t value where the ray (C,D) hits the plane N. */
			float plane_intersect(vec3 C,vec3 D,  vec3 N) {
				float a=dot(C,N);
				float b=dot(D,N);
				float t=-a/b;
				return t;
			}
			
			void main(void) {
				vec3 C=camera; // origin of ray (world coords)
				vec3 D=location-camera; // direction of ray (world coords)
			
			/* Object info */
				vec3 N1=normalize(vec3(1.0,0.0,1.0)); // normal of plane (world coordinates)
				vec4 C1=vec4(0.8,0.0,0.2,1.0); // color of plane
				
				vec3 N2=normalize(vec3(-1.0,0.0,1.0)); // normal of plane (world coordinates)
				vec4 C2=vec4(0.2,0.0,0.8,1.0); // color of plane
				
			/* Do all ray-object intersections */
				/* This is the location, normal, and color of the object we hit. */
				float t=1.0e9;
				vec3 P=vec3(0.0); 
				vec3 N=vec3(0.0); 
				vec4 color=vec4(0.0); /* transparent */
			
				// test against object 1
				float t1=plane_intersect(C,D, N1);
				vec3 P1=C+t1*D; // ray-object intersection point (world coordinates)
				if (t1<t && t1>0.0 && length(P1)<1.2) { // we have a valid hit
					t=t1; P=P1; N=N1; color=C1;
				}
				
				// test against object 2
				float t2=plane_intersect(C,D, N2);
				vec3 P2=C+t2*D; // ray-object intersection point (world coordinates)
				if (t2<t && t2>0.0 && length(P2)<1.4) { // we have a valid hit
					t=t2; P=P2; N=N2; color=C2;
				}
				
			/* Do lighting */
				vec3 L=normalize(vec3(0.3,-0.5,1.0));
				vec3 H=normalize(L+normalize(-D));
				float specular=pow(clamp(dot(H,N),0.0,1.0),500.0);
				
				gl_FragColor.rgb=
					clamp(dot(N,L),0.0,1.0)*color+
					specular*vec3(1.0,1.0,1.0);
				gl_FragColor.a=color.a; // opaque
				
			/* Compute depth */
				vec4 projP = gl_ProjectionMatrix*vec4(P,1);
				float depth = projP.z/projP.w; // perspective divide
				gl_FragDepth = 0.5+0.5 * depth; // compensate for glDepthRange
			}
		);
		
		glUseProgramObjectARB(prog);
		unsigned int ul=glGetUniformLocationARB(prog,"camera");
		glUniform3fvARB(ul,1,camera);

		/* Draw raytracer proxy geometry */
		glColor4f(1.0f,1.0f,1.0f,0.4f);
		glutSolidSphere(1.9,4,8);
		
		glUseProgramObjectARB(0);
		
		/* Draw non-raytraced geometry */
		glColor4f(0.0f,1.0f,0.0f,0.4f);
		glutSolidSphere(0.7,40,80);
	}
};


/* Called create a new simulation */
void physics_setup(physics::library &lib) {
	lib.world->add(new rayObject());
	lib.background[0]=lib.background[1]=0.6;
	lib.background[2]=1.0; // light blue
}

/* Called every frame to draw everything */
void physics_draw_frame(physics::library &lib) {
	/* Point and line smoothing needs usual alpha blending */
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
	/* Draw opaque white ground plane */
	float groundz=-1.0;
	float groundsize=30;
	glEnable(GL_DEPTH_TEST);
	glColor3f(1,1,1);
	glBegin(GL_QUADS);
	glVertex3f(-groundsize,-groundsize,groundz);
	glVertex3f(+groundsize,-groundsize,groundz);
	glVertex3f(+groundsize,+groundsize,groundz);
	glVertex3f(-groundsize,+groundsize,groundz);
	glEnd();
	
	/* No depth while drawing ground grid */
	glDisable(GL_DEPTH_TEST);
	glColor4f(0.0,0.0,0.0,1.0);
	glEnable(GL_LINE_SMOOTH);
	groundz-=0.001;
	glLineWidth(2.0);
	if (camera.z>=groundz) { /* no grid if you're underneath ground */
		glBegin(GL_LINES);
		for (float grid=-groundsize;grid<=+groundsize;grid++) {
			glVertex3f(-groundsize,grid,groundz); 
			glVertex3f(+groundsize,grid,groundz); 
			glVertex3f(grid,-groundsize,groundz); 
			glVertex3f(grid,+groundsize,groundz); 
		}
		glEnd();
	}
	
	/* Do depth while drawing objects */
	glEnable(GL_DEPTH_TEST);
	lib.world->draw(lib);
	
	
}


